/*

HXED, Version 2.00, a Hexadecimal Editor for Binary Files
Copyright 1991 Anthony F. Stuart, All rights reserved.

Date    Name	Change
910209	AFS	Initial release

*/


#include <sys/types.h>

#include "sd.h"
#include "def.h"
#include "ext.h"
#include "mcb.h"
#include "proto.h"

extern unsigned char asc_map[];
extern unsigned char hex_map[];

/*

   A note about keys... This program uses the MS-DOS model to
   represent a key. Each key is a 2 byte quantity. The low order
   byte is the ascii value. The high order byte is a value that
   uniquely identifies the specific key or key sequence. Under
   MS-DOS this is the keyboard scan code. The low order byte is
   zero for non-ascii keys, such as Alt-S. These are referred
   to in this program as command keys. 

   Command keys that are read by the ascii or hex input 
   routines are passed back to the caller and eventually handed
   off to the process_cmd_key function.

*/

/* Read an ascii value, store it, and update display */

get_ascii_input ()
{
	int	 chr;
	int	 key;
	int	 offset;
	int	 row_offset;
	int	 col_offset;

	/* Convert this addr into an offset into the screen */

	offset = this_addr - base_addr;
	row_offset = text_row + (offset / bytes_per_line);
	col_offset = text_col + (offset % bytes_per_line);

	/* Position cursor to ascii display area */

	position_cursor (row_offset, asc_map[col_offset]);

	/* Get a key, can be either ascii or command key, e.g. alt-s */

	key = get_keystroke (CMD_OR_DATA);
	chr = key & 0xff;
	if (!chr) 
		return (key >> 8);

	/* Store the new character in the file and update the display */

	store_data (chr);
	return (RIGHT);
}

/* Read a hex value, store it, and update display */

get_hex_input ()
{
	x_char	 old_value;
	x_char	 new_value;
	int	 key;

	/* Get the current value at this address */

	if (this_addr >= mcb->len)
		old_value = mcb->def_fill;
	else
		old_value = get_byte (this_addr);

	/* Get a new value or command key and update if different */

	get_hex_value (old_value, &new_value, &key);
	if (new_value != old_value)
		store_data (new_value);

	return (key);
}

/* Read a one or two byte hex sequence or command key */

get_hex_value (old_value, new_value_ptr, key_ptr)
x_char	 old_value;
x_char	*new_value_ptr;
int	*key_ptr;
{
	int	 offset;
	int	 col_offset;
	int	 row;
	int	 col;

	int	 key;
	int	 digit;
	int	 left_digit;
	int	 right_digit;

	/* Convert this addr into an offset into the screen */

	offset = this_addr - base_addr;
	col_offset = (offset % bytes_per_line);
	col = text_col + hex_map[col_offset];
	row = text_row + (offset / bytes_per_line);

	/* Extract the left and right nibbles of the current byte */

	left_digit = old_value >> 4;
	right_digit = old_value & 0x0f;

	/* Process input until a valid byte or command key is read */

	for (;;) {

		/* Get the left digit */

		get_hex_digit (row, col, &key, &digit);
		if (key) {
			if (key != RIGHT) {
				*key_ptr = key;
				*new_value_ptr = make_byte (left_digit, right_digit);
				return;
			}
		}
		else {
			left_digit = digit;
		}

		/* Get the right digit */

		get_hex_digit (row, col+1, &key, &digit);
		if (key) {
			if (key != LEFT) {
				*key_ptr = key;
				*new_value_ptr = make_byte (left_digit, right_digit);
				return;
			}
		}
		else {
			*key_ptr = RIGHT;
			*new_value_ptr = make_byte (left_digit, digit);
			return;
		}
	}
}

/* Position the cursor and read a single hex digit or command key */

get_hex_digit (row, col, key_ptr, digit_ptr)
int	 row;
int	 col;
int	*key_ptr;
int	*digit_ptr;
{
	int	 chr;
	int	 key;
	int	 digit;

	position_cursor (row, col);

	for (;;) {

		/* Wait for keystroke and get scan code and char value */

		key = get_keystroke (CMD_OR_DATA);
		chr = key & 0xff;

		/* Return scan code if command key */

		if (chr == 0) {
			*key_ptr = key >> 8;
			*digit_ptr = 0;
			return;
		}

		/* If input was a hex digit display and return it */

		digit = valid_hex_digit (chr);
		if (digit != ERR) {
			display_char (row, col, chr, cursor_attr);
			*key_ptr = 0;
			*digit_ptr = digit;
			return;
		}

		/* Otherwise let user know input was invalid and retry */

		beep ();
	}
}

/* Read a line of text. Handle line editing */

get_text_input (prompt, buf, len)
char	*prompt;
char	*buf;
int	 len;
{
	int	 chr;
	int	 key;
	int	 buf_len;
	int	 prompt_len;
	int	 curcol;
	int	 done;

	char	*bp;
	char	*bp_max;

	/* Determine optional prompt and preinitialized buffer length */

	prompt_len = strlen (prompt);
	buf_len = strlen (buf);
	bp_max = buf + len;
	bp = buf + buf_len;
	curcol = footer_col + prompt_len + buf_len;

	/* Display prompt if one was specified */

	if (prompt_len)
		display_string (footer_row, footer_col, prompt, footer_attr);

	/* Display preinitialized buffer if anything was specified */

	if (buf_len)
		display_string (footer_row, footer_col + prompt_len, buf, footer_attr);

	/* Truncate anything else on this row */

	truncate_row (footer_row, curcol, footer_attr);
	done = FALSE;

	while (!done) {

		/* Position the cursor and get a keystroke */

		position_cursor (footer_row, curcol);
		key = get_keystroke (DATA);
		chr = key & 0xff;

		/* On backspace erase previous character if one was there */

		if (chr == '\b') {
			if (bp > buf) {
				bp--;
				display_char (footer_row, --curcol, ' ', footer_attr);
			}
			else {
				beep ();
			}
		}

		/* On escape erase everything in the buffer */

		else if (chr == '\033') {
			while (bp > buf) {
				bp--;
				display_char (footer_row, --curcol, ' ', footer_attr);
			}
		}

		/* Terminate input on line feed or carriage return */

		else if (chr == '\n' || chr == '\r') {
			*bp = EOS;
			done = TRUE;
		}

		/* Store character if just normal ascii */

		else if (chr) {
			if ((bp < bp_max) && (curcol < maxcol)) {
				*bp++ = chr;
				display_char (footer_row, curcol++, chr, footer_attr);
			}
			else {
				beep ();
			}
		}

		/* Else beep, this function doesn't support command keys */

		else {
			beep ();
		}
	}

	/* After input is terminated, redisplay footer and home cursor */

	display_footer ();
	position_cursor (minrow, mincol);
}

/* Convert escape sequences to corresponding ascii or 8 bit values */

/* Return addr of next position in out_buf */
/* Because nulls may occur in text stream */

x_char *
convert_esc_seq (in_buf, out_buf)
x_char	*in_buf;
x_char	*out_buf;
{
	while (*in_buf) {
		if (*in_buf == '\\') {
			switch (*(++in_buf)) {
				case 'a':
					*out_buf++ = '\a';
					break;
				case 'b':
					*out_buf++ = '\b';
					break;
				case 'f':
					*out_buf++ = FF;
					break;
				case 'n':
					*out_buf++ = '\n';
					break;
				case 'r':
					*out_buf++ = '\r';
					break;
				case 't':
					*out_buf++ = '\t';
					break;
				case 'v':
					*out_buf++ = VT;
					break;
				case '*':
					*out_buf++ = '\\';
					*out_buf++ = '*';
					break;
				case '?':
					*out_buf++ = '\\';
					*out_buf++ = '?';
					break;
				case 'd':
					*out_buf++ = convert_decimal (&in_buf);
					break;
				case 'x':
					*out_buf++ = convert_hex (&in_buf);
					break;
				case '0':
					*out_buf++ = convert_octal (&in_buf);
					break;
				default:
					*out_buf++ = *in_buf;
					break;
			}
		}
		else
			*out_buf++ = *in_buf;

		in_buf++;
	}
	return (out_buf);
}

/* Convert octal number from string to integer */
/* representation so it can be stored as a single byte. */
/* Return first non-octal character via parameter list. */

convert_octal (in_buf)
char	**in_buf;
{
	char	*bp;
	int	 value;

	/* Initialize to first character in buffer and zero value */

	bp = *in_buf;
	value = 0;

	/* Convert any octal characters to integer */
	/* Return decimal integer and address of first non dec character */


	while (*bp >= '0' && *bp <= '7') {
		value = (value << 3) + (*bp - '0');
		bp++;
	}

	/* Return integer and address of first non octal character */

	*in_buf = bp - 1;
	return (value);
}

/* Convert decimal number from string to integer */
/* representation so it can be stored as a single byte. */
/* Return first non-decimal character via parameter list. */

convert_decimal (in_buf)
char	**in_buf;
{
	char	*bp;
	int	 value;

	/* Initialize to first character following \d */

	bp = (*in_buf)+1;
	value = 0;

	/* Convert any decimal characters to integer */

	while (*bp >= '0' && *bp <= '9') {
		value = (value * 10) + (*bp - '0');
		bp++;
	}

	/* Return integer and address of first non decimal character */

	*in_buf = bp - 1;
	return (value);
}

/* Convert hexadecimal number from string to integer */
/* representation so it can be stored as a single byte. */
/* Return first non-hex character via parameter list. */

convert_hex (in_buf)
char	**in_buf;
{
	int	 c;
	char	*bp;
	int	 value;

	/* Initialize to first character following \x */

	bp = (*in_buf)+1;
	value = 0;

	/* Convert any hexadecimal characters to integer */

	while ((c = valid_hex_digit (*bp)) != ERR) {
		value = (value << 4) + c;
		bp++;
	}

	/* Return integer and address of first non hexadecimal character */

	*in_buf = bp - 1;
 	return (value);
}

/* Return 0-15 if c is a valid hex digit or ERR if not */

valid_hex_digit (c)
{
	char	*bp;

	for (bp = lookup; *bp; bp++) {
		if (*bp == c)
			return (bp - lookup);
	}
	return (ERR);
}

/* Prompt for and get a valid response */
/* Note: only first character is validated */

get_response (prompt, valid)
char	*prompt;		
char	*valid;
{
	char	*vp;
	char	 buf[BUF_SZ];

	for (;;) {

		/* Write prompt and get input */

		buf[0] = EOS;
		get_text_input (prompt, buf, BUF_SZ);

		/* Convert uppercase to lowercase */

		if (buf[0] >= 'A' && buf[0] <= 'Z')
			buf[0] += 0x20;

		/* Return if in array of valid responses */

		for (vp = valid; *vp; vp++) {
			if (buf[0] == *vp)
				return (*vp);
		}

		/* Otherwise try again */

		display_message ("Invalid response, try again");
	}
}


